/**
 * @file sid_epoc.i - 6581 emulation, EPOC specific stuff
 *
 * (c) 2001-2002 Alfred E. Heggestad
 */

#ifndef __ER6__
#include "alaw.h"
#endif

// #define PCM_DUMP

/**
 * Buffer size: 2^9 == 512 bytes. Note that too large buffers will not work
 * very well: The speed of the C64 is slowed down to an average speed of 
 * 100% by the blocking write() call in EmulateLine(). If you use a buffer 
 * of, say 4096 bytes, that will happen only about every 4 frames, which
 * means that the emulation runs much faster in some frames, and much
 * slower in others.
 * On really fast machines, it might make sense to use an even smaller
 * buffer size.
 */
const TInt KDefaultBufSize = 512;


void DigitalRenderer::init_sound()
/**
 *  Initialization
 */
	{
#ifdef __ER6__

	/// @todo sound for ER6

    ready = false; // asume the worst

#else // __ER6__
    divisor = to_output = buffer_pos = 0;

    ready = false; // asume the worst

	/**
	 * open sound device
	 */
	TInt ret = iDevSound.Open();
	if (ret != KErrNone)
		{
		ELOG2(_L8("AUDIO: could not open sound device [ret=%d]\n"), ret);
		return;
		}

	/**
	 * get sound device capabilites
	 */ 
	TSoundCaps sndCaps;
	iDevSound.Caps(sndCaps);
	ELOG1(_L8("AUDIO: RDevSound - caps:\n"));
	ELOG2(_L8("  iService      = 0x%08x\n"), sndCaps().iService);
	ELOG2(_L8("  iVolume       = 0x%08x\n"), sndCaps().iVolume);
	ELOG2(_L8("  iMaxVolume    = 0x%08x\n"), sndCaps().iMaxVolume);
	ELOG2(_L8("  iMaxFrequency = 0x%08x\n"), sndCaps().iMaxFrequency);

	/**
	 * configure sound device
	 */
	TSoundConfig sndConfig;
	iDevSound.Config(sndConfig); // read the default config

	sndConfig().iVolume = EVolumeByValue; // 4
//	sndConfig().iVolumeValue = sndCaps().iMaxVolume; @todo
	sndConfig().iVolumeValue = 1;
	sndConfig().iAlawBufferSize = KDefaultBufSize * 2; // double buffering
	
	ret = iDevSound.SetConfig(sndConfig);
	if(ret != KErrNone)
		{
		ELOG2(_L8("AUDIO: WARNING! Could not set config due to [%d]\n"), ret);
		iDevSound.Close();
		return;
		}

	ELOG2(_L8("AUDIO: iVolume         %d \n"), sndConfig().iVolume );
	ELOG2(_L8("       iAlawBufferSize %d \n"), sndConfig().iAlawBufferSize );

	/**
	 * prepare the buffer
	 */
	ret = iDevSound.PreparePlayAlawBuffer(); // This *must* be done before reading the Buffer size!!!
	if(ret != KErrNone)
		{
		ELOG2(_L8("AUDIO: WARNING! could not prepare alaw buffer due to [%d]\n"), ret);
		return;
		}

	sound_buffer = new int16[KDefaultBufSize];
	iAlawBuffer = HBufC8::New(KDefaultBufSize);
	if(!sound_buffer || !iAlawBuffer)
		{
		ELOG1(_L8("AUDIO: no memory left for audio buffers :(\n"));
		iDevSound.Close();
		return;
		}

	ELOG1(_L8("AUDIO: the SID chip is ready to Rock'n'Roll...\n"));
    ready = true;

#endif // __ER6__
	}


DigitalRenderer::~DigitalRenderer()
/**
 *  Destructor
 */
	{
#ifndef __ER6__
	delete iAlawBuffer;
	delete sound_buffer;
	iDevSound.Close();
#endif
	}


void DigitalRenderer::Pause(void)
/**
 *  Pause sound output (do nothing)
 */
	{
	}


void DigitalRenderer::Resume(void)
/**
 * Resume sound output (do nothing)
 */
	{
	}


void DigitalRenderer::EmulateLine(void)
/**
 * Fill buffer, sample volume (for sampled voice)
 */
	{
	if (!ready)
		return;

	sample_buf[sample_in_ptr] = volume;
	sample_in_ptr = (sample_in_ptr + 1) % SAMPLE_BUF_SIZE;

    /**
     * Now see how many samples have to be added for this line
     */
    divisor += SAMPLE_FREQ;
    while (divisor >= 0)
		divisor -= TOTAL_RASTERS*SCREEN_FREQ, to_output++;

    /**
     * Calculate the sound data only when we have enough to fill
     * the buffer entirely.
     */
    if ((buffer_pos + to_output) >= KDefaultBufSize)
		{
		int datalen = KDefaultBufSize - buffer_pos;
		to_output -= datalen;
		calc_buffer(sound_buffer + buffer_pos, datalen*2);

#ifdef PCM_DUMP

		_LIT(KPcmDumpFile, "c:\\e32frodo_pcm.dat");
		const TInt KFileOpenWritable = EFileWrite | EFileShareAny | EFileStream;
		
		RFs fs;
		TInt ret = fs.Connect();
		if( (ret == KErrNone) || (ret == KErrAlreadyExists) )
			{
			RFile file;
			ret = file.Open(fs, KPcmDumpFile, KFileOpenWritable);
			if(ret == KErrNotFound)
				ret = file.Replace(fs, KPcmDumpFile, KFileOpenWritable);
			if(ret == KErrNone)
				{
				TInt pos;
				ret = file.Seek(ESeekEnd, pos);
				if (ret==KErrNone)
					{
					TPtrC8 ptr((unsigned char*)sound_buffer, KDefaultBufSize*2); // 2x8 = 16
					file.Write(ptr);
					}
				else
					{
					ELOG2(_L8("PCMDUMP: could not seek due to %d\n"), ret);
					}
				
				file.Close();
				}
			else
				{
				ELOG2(_L8("PCMDUMP: could not open file due to %d\n"), ret);
				}
			
			fs.Close();
			}

#else // PCM_DUMP

#ifdef __ER6__

		/// @todo add sound for ER6 here

#else // __ER6__

		//
		// convert from linear PCM to aLaw first
		//
		TPtr8 alawPtr = iAlawBuffer->Des();
		alawPtr.SetLength(KDefaultBufSize);
		Alaw::conv_s16bit_alaw((unsigned short*)sound_buffer, &alawPtr[0], KDefaultBufSize);
		
		//
		// and then play the Alaw buffer
		//
		TRequestStatus stat;
		iDevSound.PlayAlawData(stat, alawPtr);
		User::WaitForRequest(stat);
		TInt ret = stat.Int();
		if(ret != KErrNone)
			{
			ELOG2(_L8("AUDIO: Warning! could not play due to %d\n"), ret);
			}
#endif // __ER6__

#endif // PCM_DUMP

//
// this is how they do it on Linux:
//
//		write(devfd, sound_buffer, sndbufsize*2);

		buffer_pos = 0;
		}
    
	}

//
// this code is moved here from FixPoint.i to avoid global data
//

/// define as global variable (4x512=2048 bytes)
//FixPoint SinTable[(1<<ldSINTAB)];


#define FIXPOINT_SIN_COS_GENERIC \
  if (angle >= 3*(1<<ldSINTAB)) {return(-SinTable[(1<<(ldSINTAB+2)) - angle]);}\
  if (angle >= 2*(1<<ldSINTAB)) {return(-SinTable[angle - 2*(1<<ldSINTAB)]);}\
  if (angle >= (1<<ldSINTAB)) {return(SinTable[2*(1<<ldSINTAB) - angle]);}\
  return(SinTable[angle]);


// sin and cos: angle is fixpoint number 0 <= angle <= 2 (*PI)
inline FixPoint DigitalRenderer::fixsin(FixPoint x)
{
  int angle = x;

  angle = (angle >> (FIXPOINT_PREC - ldSINTAB - 1)) & ((1<<(ldSINTAB+2))-1);
  FIXPOINT_SIN_COS_GENERIC
}


inline FixPoint DigitalRenderer::fixcos(FixPoint x)
{
  int angle = x;

  // cos(x) = sin(x+PI/2)
  angle = (angle + (1<<(FIXPOINT_PREC-1)) >> (FIXPOINT_PREC - ldSINTAB - 1)) & ((1<<(ldSINTAB+2))-1);
  FIXPOINT_SIN_COS_GENERIC
}


inline void DigitalRenderer::InitFixSinTab(void)
{
 ELOG2(_L8("InitFixSinTab [SinTable=0x%08x] ..."), SinTable);
  int i;
  float step;

  for (i=0, step=0; i<(1<<ldSINTAB); i++, step+=0.5/(1<<ldSINTAB))
  {
    SinTable[i] = FixNo(sin(M_PI * step));
  }
  ELOG1(_L8("[  OK  ]\n"));
}
